﻿#include "noun.h"

#include <iostream> //Xperimental -- for debug only

#include "utf8utf16.h"

#include "boost/regex.hpp"
#include "boost/algorithm/string/regex.hpp"


std::vector<NounRecord> NounRecordArr;

std::vector<NounDeclencionCaseTableRecord> nounDeclencionCaseTable;

NounRecord::NounRecord()
	: gender(NG_MALE)
	, haveSingleForm(false)
	, haveMultipleForm(false)
	, haveStandardMultipleForm(false)
	, haveStandardMultipleFormWithMissingLastVowel(false)
	, canBeAnimate(false)
	, canBeInanimate(false)
{

}


NounRecord::NounRecord(std::wstring line)
{
	std::vector<std::wstring> lineArr;

	boost::split_regex(lineArr, line, boost::wregex(L";"));

	nominativeForm = lineArr[1];

	if (lineArr[2] == L"м")
	{
		gender = NG_MALE;
	}
	else if (lineArr[2] == L"ж")
	{
		gender = NG_FEMALE;
	}
	else
	{
		gender = NG_NEUTRAL;
	}

	haveSingleForm = lineArr[3] == L"1" ? true : false;
	haveMultipleForm = lineArr[4] == L"1" ? true : false;

	haveStandardMultipleForm = lineArr[5] == L"1" ? true : false;

	haveStandardMultipleFormWithMissingLastVowel = lineArr[6] == L"1" ? true : false;
	
	
	specialMultipleForm = lineArr[7];

	canBeAnimate = lineArr[8] == L"1" ? true : false;
	canBeInanimate = lineArr[9] == L"1" ? true : false;


}


NounDeclencion WStringToNounDeclencion(std::wstring str)
{

	if (str == L"FIRST_A_IFORM_SINGULAR")
	{
		return FIRST_A_IFORM_SINGULAR;
	}
	if (str == L"FIRST_A_IFORM_INANIMATE_PLURAL")
	{
		return FIRST_A_IFORM_INANIMATE_PLURAL;
	}
	if (str == L"FIRST_A_IFORM_ANIMATE_PLURAL")
	{
		return FIRST_A_IFORM_ANIMATE_PLURAL;
	}
	if (str == L"FIRST_A_UFORM_SINGULAR")
	{
		return FIRST_A_UFORM_SINGULAR;
	}
	if (str == L"FIRST_A_UFORM_INANIMATE_PLURAL")
	{
		return FIRST_A_UFORM_INANIMATE_PLURAL;
	}
	if (str == L"FIRST_A_UFORM_ANIMATE_PLURAL")
	{
		return FIRST_A_UFORM_ANIMATE_PLURAL;
	}
	if (str == L"FIRST_YA_FORM_SINGULAR")
	{
		return FIRST_YA_FORM_SINGULAR;
	}
	if (str == L"FIRST_YA_FORM_INANIMATE_PLURAL")
	{
		return FIRST_YA_FORM_INANIMATE_PLURAL;
	}
	if (str == L"FIRST_YA_FORM_ANIMATE_PLURAL")
	{
		return FIRST_YA_FORM_ANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_IFORM_INANIMATE_SINGULAR")
	{
		return SECOND_MALE_IFORM_INANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_IFORM_INANIMATE_PLURAL")
	{
		return SECOND_MALE_IFORM_INANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_IFORM_ANIMATE_SINGULAR")
	{
		return SECOND_MALE_IFORM_ANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_IFORM_ANIMATE_PLURAL")
	{
		return SECOND_MALE_IFORM_ANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_UFORM_INANIMATE_SINGULAR")
	{
		return SECOND_MALE_UFORM_INANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_UFORM_INANIMATE_PLURAL")
	{
		return SECOND_MALE_UFORM_INANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_UFORM_ANIMATE_SINGULAR")
	{
		return SECOND_MALE_UFORM_ANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_UFORM_ANIMATE_PLURAL")
	{
		return SECOND_MALE_UFORM_ANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_SSFORM_INANIMATE_SINGULAR")
	{
		return SECOND_MALE_SSFORM_INANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_SSFORM_INANIMATE_PLURAL")
	{
		return SECOND_MALE_SSFORM_INANIMATE_PLURAL;
	}
	if (str == L"SECOND_MALE_SSFORM_ANIMATE_SINGULAR")
	{
		return SECOND_MALE_SSFORM_ANIMATE_SINGULAR;
	}
	if (str == L"SECOND_MALE_SSFORM_ANIMATE_PLURAL")
	{
		return SECOND_MALE_SSFORM_ANIMATE_PLURAL;
	}
	if (str == L"SECOND_NEUTRAL_E_FORM_SINGULAR")
	{
		return SECOND_NEUTRAL_E_FORM_SINGULAR;
	}
	if (str == L"SECOND_NEUTRAL_E_FORM_PLURAL")
	{
		return SECOND_NEUTRAL_E_FORM_PLURAL;
	}
	if (str == L"SECOND_NEUTRAL_O_FORM_SINGULAR")
	{
		return SECOND_NEUTRAL_O_FORM_SINGULAR;
	}
	if (str == L"SECOND_NEUTRAL_O_FORM_PLURAL")
	{
		return SECOND_NEUTRAL_O_FORM_PLURAL;
	}
	if (str == L"THIRD_FORM_SINGULAR")
	{
		return THIRD_FORM_SINGULAR;
	}
	if (str == L"THIRD_FORM_INANIMATE_PLURAL")
	{
		return THIRD_FORM_INANIMATE_PLURAL;
	}
	if (str == L"THIRD_FORM_ANIMATE_PLURAL")
	{
		return THIRD_FORM_ANIMATE_PLURAL;
	}

	std::cout << "Error in WStringToNounDeclencion!" << std::endl;
	return FIRST_A_IFORM_SINGULAR;

}


std::wstring NounGrammaticalCaseToWString(NounGrammaticalCase nounGrammaticalCase)
{
	switch (nounGrammaticalCase)
	{
	case NGC_P1_NOMINATIVE: return L"NGC_P1_NOMINATIVE";
	case NGC_P2_GENITIVE: return L"NGC_P2_GENITIVE";
	case NGC_P3_DATIVE: return L"NGC_P3_DATIVE";
	case NGC_P4_ACCUSATIVE: return L"NGC_P4_ACCUSATIVE";
	case NGC_P5_INSTRUMENTAL: return L"NGC_P5_INSTRUMENTAL";
	case NGC_P6_PREPOSITIONAL: return L"NGC_P6_PREPOSITIONAL";
	}

	return L"";
}

NounGrammaticalCase WStringToNounGrammaticalCase(std::wstring str)
{

	if (str == L"NGC_P1_NOMINATIVE")
	{
		return NGC_P1_NOMINATIVE;
	}
	if (str == L"NGC_P2_GENITIVE")
	{
		return NGC_P2_GENITIVE;
	}
	if (str == L"NGC_P3_DATIVE")
	{
		return NGC_P3_DATIVE;
	}
	if (str == L"NGC_P4_ACCUSATIVE")
	{
		return NGC_P4_ACCUSATIVE;
	}
	if (str == L"NGC_P5_INSTRUMENTAL")
	{
		return NGC_P5_INSTRUMENTAL;
	}
	if (str == L"NGC_P6_PREPOSITIONAL")
	{
		return NGC_P6_PREPOSITIONAL;
	}

	std::cout << "Error in WStringToNounGrammaticalCase!" << std::endl;
	return NGC_P1_NOMINATIVE;
}

std::vector<std::wstring> GetAllNounEndingArr()
{
	std::vector<std::wstring> result
	{
		L"",
		L"а",
		L"и",
		L"е",
		L"у",
		L"ой",
		L"ы",
		L"ом",
		L"ь",
		L"я",
		L"ю",
		L"ем",
		L"о",
		L"ью",
		L"ам",
		L"ами",
		L"ах",
		L"ов",
		L"ей",
		L"ям",
		L"ях",
		L"я",
		L"ями",
	};

	return result;
}

bool NounIsInDictionary(std::wstring nounNominative)
{
	for (auto& noun : NounRecordArr)
	{
		if (noun.nominativeForm == nounNominative)
		{
			return true;
		}
	}

	return false;

}

bool NounSpecialPluralFormIsInDictionary(std::wstring nounNominativePlural)
{
	for (auto& noun : NounRecordArr)
	{
		if (noun.specialMultipleForm == nounNominativePlural)
		{
			return true;
		}
	}

	return false;
}

NounRecord GetNounRecordFromDictionary(std::wstring nounNominative)
{
	for (auto& noun : NounRecordArr)
	{
		if (noun.nominativeForm == nounNominative)
		{
			return noun;
		}
	}

	return{};

}

NounRecord GetNounRecordFromDictionary_BySpecialPluralForm(std::wstring nounNominativePlural)
{
	for (auto& noun : NounRecordArr)
	{
		if (noun.specialMultipleForm == nounNominativePlural)
		{
			return noun;
		}
	}

	return{};

}


bool charIsConsolant(wchar_t c)
{
	std::wstring consolants = L"йцкнгшщзхфвпрлджчсмтб";

	for (wchar_t ic : consolants)
	{
		if (c == ic)
		{
			return true;
		}
	}

	return false;
}

bool charIsVowel(wchar_t c)
{
	std::wstring vovels = L"аоуыэяёюие";

	for (wchar_t ic : vovels)
	{
		if (c == ic)
		{
			return true;
		}
	}

	return false;
}


std::vector<std::pair<std::wstring, std::wstring>> getPossibleNounEndingDivisionArr(std::wstring noun)
{
	std::vector<std::pair<std::wstring, std::wstring>> result;

	auto allNounEndingArr = GetAllNounEndingArr();

	for (auto ending : allNounEndingArr)
	{
		if (boost::ends_with(noun, ending))
		{
			std::wstring nounBase = boost::replace_last_copy(noun, ending, "");

			if (charIsConsolant(nounBase[nounBase.size() - 1]) || nounBase[nounBase.size() - 1] == L'ь' || nounBase[nounBase.size() - 1] == L'ъ')
			{
				result.push_back({ nounBase, ending});
			}
		}
	}

	return result;
}



std::vector<NounTuple> GetPossibleNounTupleArr(std::wstring nounEnding)
{
	std::vector<NounTuple> result;

	for (auto& noun : nounDeclencionCaseTable)
	{
		for (int i = 0; i < 6; i++)
		{
			if (noun.grammaticalCaseMap[i].second == nounEnding)
			{
				result.push_back(NounTuple{ noun.nounDeclencion, noun.grammaticalCaseMap[i].first });
			}
		}

	}

	return result;
}

bool NounDeclencionIsPlural(NounDeclencion nounDeclencion)
{
	switch (nounDeclencion)
	{
	case FIRST_A_IFORM_SINGULAR:
	case FIRST_A_UFORM_SINGULAR:
	case FIRST_YA_FORM_SINGULAR:
	case SECOND_MALE_IFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_IFORM_ANIMATE_SINGULAR:
	case SECOND_MALE_UFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_UFORM_ANIMATE_SINGULAR:
	case SECOND_MALE_SSFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_SSFORM_ANIMATE_SINGULAR:
	case SECOND_NEUTRAL_E_FORM_SINGULAR:
	case SECOND_NEUTRAL_O_FORM_SINGULAR:
	case THIRD_FORM_SINGULAR:
		return false;
		break;
	default:
		return true;
		break;
	}
}

NounDeclencion ConvertToSingularIfNeeded(NounDeclencion nounDeclencion)
{
	switch (nounDeclencion)
	{
	case FIRST_A_IFORM_SINGULAR:
	case FIRST_A_IFORM_INANIMATE_PLURAL:
	case FIRST_A_IFORM_ANIMATE_PLURAL:

		return FIRST_A_IFORM_SINGULAR;

		break;
	case FIRST_A_UFORM_SINGULAR:
	case FIRST_A_UFORM_INANIMATE_PLURAL:
	case FIRST_A_UFORM_ANIMATE_PLURAL:

		return FIRST_A_UFORM_SINGULAR;

		break;
	case FIRST_YA_FORM_SINGULAR:
	case FIRST_YA_FORM_INANIMATE_PLURAL:
	case FIRST_YA_FORM_ANIMATE_PLURAL:

		return FIRST_YA_FORM_SINGULAR;

		break;
	case SECOND_MALE_IFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_IFORM_INANIMATE_PLURAL:

		return SECOND_MALE_IFORM_INANIMATE_SINGULAR;

		break;
	case SECOND_MALE_IFORM_ANIMATE_SINGULAR:
	case SECOND_MALE_IFORM_ANIMATE_PLURAL:

		return SECOND_MALE_IFORM_ANIMATE_SINGULAR;

		break;
	case SECOND_MALE_UFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_UFORM_INANIMATE_PLURAL:

		return SECOND_MALE_UFORM_INANIMATE_SINGULAR;

		break;
	case SECOND_MALE_UFORM_ANIMATE_SINGULAR:
	case SECOND_MALE_UFORM_ANIMATE_PLURAL:

		return SECOND_MALE_UFORM_ANIMATE_SINGULAR;

		break;
	case SECOND_MALE_SSFORM_INANIMATE_SINGULAR:
	case SECOND_MALE_SSFORM_INANIMATE_PLURAL:

		return SECOND_MALE_SSFORM_INANIMATE_SINGULAR;

		break;
	case SECOND_MALE_SSFORM_ANIMATE_SINGULAR:
	case SECOND_MALE_SSFORM_ANIMATE_PLURAL:

		return SECOND_MALE_SSFORM_ANIMATE_SINGULAR;

		break;
	case SECOND_NEUTRAL_E_FORM_SINGULAR:
	case SECOND_NEUTRAL_E_FORM_PLURAL:

		return SECOND_NEUTRAL_E_FORM_SINGULAR;

		break;
	case SECOND_NEUTRAL_O_FORM_SINGULAR:
	case SECOND_NEUTRAL_O_FORM_PLURAL:

		return SECOND_NEUTRAL_O_FORM_SINGULAR;

		break;
	case THIRD_FORM_SINGULAR:
	case THIRD_FORM_INANIMATE_PLURAL:
	case THIRD_FORM_ANIMATE_PLURAL:

		return THIRD_FORM_SINGULAR;

		break;
	default:

		std::cout << "Error in ConvertToSingularIfNeeded" << std::endl;

		return FIRST_A_IFORM_SINGULAR;
		break;
	}
}

std::vector<NounTuple> FilterNounTupleArrByNounDeclentionSet(std::vector<NounTuple> nounTupleArr, std::set<NounDeclencion> filter)
{
	std::vector<NounTuple> result;

	for (auto nounTuple : nounTupleArr)
	{
		if (filter.count(std::get<0>(nounTuple)) != 0)
		{
			result.push_back(nounTuple);
		}
	}

	return result;
}


std::wstring GetNounNoninativeSingular(std::wstring nounBase, NounDeclencion nounDeclencion)
{
	NounDeclencion nounDeclencionSingular = ConvertToSingularIfNeeded(nounDeclencion);

	//Get nominative form of this declention
	std::wstring nounEndling = nounDeclencionCaseTable[static_cast<int>(nounDeclencionSingular)].grammaticalCaseMap[static_cast<int>(NGC_P1_NOMINATIVE)].second;

	return nounBase + nounEndling;
}


std::wstring GetNounNoninative(std::wstring nounBase, NounDeclencion nounDeclencion)
{
	std::wstring nounEndling = nounDeclencionCaseTable[static_cast<int>(nounDeclencion)].grammaticalCaseMap[static_cast<int>(NGC_P1_NOMINATIVE)].second;

	return nounBase + nounEndling;
}

std::wstring i_form_consolants = L"гкжшчщ";
std::wstring u_form_consolants = L"бпдтвфзснмлр";

wchar_t GetLastChar(const NounRecord& nounRecord)
{
	return nounRecord.nominativeForm[nounRecord.nominativeForm.size() - 1];
}

wchar_t GetPrevLastChar(const NounRecord& nounRecord)
{
	return nounRecord.nominativeForm[nounRecord.nominativeForm.size() - 2];
}

bool charIsIFormConsolant(wchar_t c)
{
	return i_form_consolants.find(c) != i_form_consolants.npos;
}

bool charIsUFormConsolant(wchar_t c)
{
	return u_form_consolants.find(c) != i_form_consolants.npos;
}


bool FirstAIFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && (GetLastChar(nounRecord) == L'а' && charIsIFormConsolant(GetPrevLastChar(nounRecord)));
}

bool FirstAIFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeInanimate && (GetLastChar(nounRecord) == L'а' && charIsIFormConsolant(GetPrevLastChar(nounRecord)));
}

bool FirstAIFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeAnimate && (GetLastChar(nounRecord) == L'а' && charIsIFormConsolant(GetPrevLastChar(nounRecord)));
}


bool FirstAUFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && (GetLastChar(nounRecord) == L'а' && charIsUFormConsolant(GetPrevLastChar(nounRecord)));
}

bool FirstAUFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeInanimate && (GetLastChar(nounRecord) == L'а' && charIsUFormConsolant(GetPrevLastChar(nounRecord)));
}

bool FirstAUFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeAnimate && (GetLastChar(nounRecord) == L'а' && charIsUFormConsolant(GetPrevLastChar(nounRecord)));
}


bool FirstYaFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && (GetLastChar(nounRecord) == L'я');
}

bool FirstYaFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeInanimate && (GetLastChar(nounRecord) == L'я');
}

bool FirstYaFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.canBeAnimate && (GetLastChar(nounRecord) == L'я');
}


// Second declention

bool SecondMaleIFormInanimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && charIsIFormConsolant(GetLastChar(nounRecord));
}

bool SecondMaleIFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && charIsIFormConsolant(GetLastChar(nounRecord));
}


bool SecondMaleIFormAnimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && charIsIFormConsolant(GetLastChar(nounRecord));
}

bool SecondMaleIFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && charIsIFormConsolant(GetLastChar(nounRecord));
}


bool SecondMaleUFormInanimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && charIsUFormConsolant(GetLastChar(nounRecord));
}

bool SecondMaleUFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && charIsUFormConsolant(GetLastChar(nounRecord));
}


bool SecondMaleUFormAnimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && charIsUFormConsolant(GetLastChar(nounRecord));
}

bool SecondMaleUFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && charIsUFormConsolant(GetLastChar(nounRecord));
}


bool SecondMaleSSFormInanimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && GetLastChar(nounRecord) == L'ь';
}

bool SecondMaleSSFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeInanimate && GetLastChar(nounRecord) == L'ь';
}


bool SecondMaleSSFormAnimateSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && GetLastChar(nounRecord) == L'ь';
}

bool SecondMaleSSFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_MALE && nounRecord.canBeAnimate && GetLastChar(nounRecord) == L'ь';
}

bool SecondNeutralEFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_NEUTRAL && GetLastChar(nounRecord) == L'е';
}

bool SecondNeutralEFormPluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_NEUTRAL && GetLastChar(nounRecord) == L'е';
}

bool SecondNeutralOFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_NEUTRAL && GetLastChar(nounRecord) == L'о';
}

bool SecondNeutralOFormPluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_NEUTRAL && GetLastChar(nounRecord) == L'о';
}

bool ThirdFormSingularCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveSingleForm && nounRecord.gender == NG_FEMALE && GetLastChar(nounRecord) == L'ь';
}

bool ThirdFormInanimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_FEMALE && nounRecord.canBeInanimate && GetLastChar(nounRecord) == L'ь';
}

bool ThirdFormAnimatePluralCondition(const NounRecord& nounRecord)
{
	return nounRecord.haveStandardMultipleForm && nounRecord.gender == NG_FEMALE && nounRecord.canBeAnimate && GetLastChar(nounRecord) == L'ь';
}


std::map<NounDeclencion, std::function < bool(const NounRecord&) >> DeclentionConditionMap;

void SetupDeclentionMap()
{

	DeclentionConditionMap[FIRST_A_IFORM_SINGULAR] = std::bind(FirstAIFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_A_IFORM_INANIMATE_PLURAL] = std::bind(FirstAIFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_A_IFORM_ANIMATE_PLURAL] = std::bind(FirstAIFormAnimatePluralCondition, std::placeholders::_1);

	DeclentionConditionMap[FIRST_A_UFORM_SINGULAR] = std::bind(FirstAUFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_A_UFORM_INANIMATE_PLURAL] = std::bind(FirstAUFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_A_UFORM_ANIMATE_PLURAL] = std::bind(FirstAUFormAnimatePluralCondition, std::placeholders::_1);

	DeclentionConditionMap[FIRST_YA_FORM_SINGULAR] = std::bind(FirstYaFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_YA_FORM_INANIMATE_PLURAL] = std::bind(FirstYaFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[FIRST_YA_FORM_ANIMATE_PLURAL] = std::bind(FirstYaFormAnimatePluralCondition, std::placeholders::_1);

	//Second form
	DeclentionConditionMap[SECOND_MALE_IFORM_INANIMATE_SINGULAR] = std::bind(SecondMaleIFormInanimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_IFORM_INANIMATE_PLURAL] = std::bind(SecondMaleIFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_IFORM_ANIMATE_SINGULAR] = std::bind(SecondMaleIFormAnimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_IFORM_ANIMATE_PLURAL] = std::bind(SecondMaleIFormAnimatePluralCondition, std::placeholders::_1);

	DeclentionConditionMap[SECOND_MALE_UFORM_INANIMATE_SINGULAR] = std::bind(SecondMaleUFormInanimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_UFORM_INANIMATE_PLURAL] = std::bind(SecondMaleUFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_UFORM_ANIMATE_SINGULAR] = std::bind(SecondMaleUFormAnimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_UFORM_ANIMATE_PLURAL] = std::bind(SecondMaleUFormAnimatePluralCondition, std::placeholders::_1);

	DeclentionConditionMap[SECOND_MALE_SSFORM_INANIMATE_SINGULAR] = std::bind(SecondMaleSSFormInanimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_SSFORM_INANIMATE_PLURAL] = std::bind(SecondMaleSSFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_SSFORM_ANIMATE_SINGULAR] = std::bind(SecondMaleSSFormAnimateSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_MALE_SSFORM_ANIMATE_PLURAL] = std::bind(SecondMaleSSFormAnimatePluralCondition, std::placeholders::_1);

	DeclentionConditionMap[SECOND_NEUTRAL_E_FORM_SINGULAR] = std::bind(SecondNeutralEFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_NEUTRAL_E_FORM_PLURAL] = std::bind(SecondNeutralEFormPluralCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_NEUTRAL_O_FORM_SINGULAR] = std::bind(SecondNeutralOFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[SECOND_NEUTRAL_O_FORM_PLURAL] = std::bind(SecondNeutralOFormPluralCondition, std::placeholders::_1);


	DeclentionConditionMap[THIRD_FORM_SINGULAR] = std::bind(ThirdFormSingularCondition, std::placeholders::_1);
	DeclentionConditionMap[THIRD_FORM_INANIMATE_PLURAL] = std::bind(ThirdFormInanimatePluralCondition, std::placeholders::_1);
	DeclentionConditionMap[THIRD_FORM_ANIMATE_PLURAL] = std::bind(ThirdFormAnimatePluralCondition, std::placeholders::_1);


}


bool NounFitsDeclention(NounRecord nounRecord, NounTuple nounTuple)
{

	NounDeclencion nounDeclencion = std::get<0>(nounTuple);

	return DeclentionConditionMap[nounDeclencion](nounRecord);
}

bool NounScructIsAlreadyInArray(const NounStruct& nounStruct, const std::vector<NounStruct>& arr)
{
	for (auto& ns : arr)
	{
		if (ns.nounGrammaticalCase == nounStruct.nounGrammaticalCase &&
			ns.nounRecord.nominativeForm == nounStruct.nounRecord.nominativeForm)
		{
			return true;
		}
	}

	return false;
}

std::vector<NounStruct> RecognizeNoun(std::wstring noun)
{

	std::vector<NounStruct> result;

	auto nounEndingDivisionArr = getPossibleNounEndingDivisionArr(noun);

	for (auto nounEndingDivision : nounEndingDivisionArr)
	{
		std::wstring nounBase = nounEndingDivision.first;
		std::wstring nounEnding = nounEndingDivision.second;

		std::vector<NounTuple> possibleTupleArr = GetPossibleNounTupleArr(nounEnding);

		//std::vector<NounTuple> sureTupleArr;
		
		//Standard check
		for (auto nounTuple : possibleTupleArr)
		{
			std::wstring nounNominaviteSingular = GetNounNoninativeSingular(nounBase, std::get<0>(nounTuple));

			if (NounIsInDictionary(nounNominaviteSingular))
			{

				NounRecord nounRecord = GetNounRecordFromDictionary(nounNominaviteSingular);

				if (NounFitsDeclention(nounRecord, nounTuple))
				{
					//sureTupleArr.push_back(nounTuple);
					result.push_back({ std::get<1>(nounTuple), nounRecord });
				}
			}
		}

		//Special plural form check
		for (auto nounTuple : possibleTupleArr)
		{
			if (NounDeclencionIsPlural(std::get<0>(nounTuple)))
			{
				std::wstring nounNominavitePlural = GetNounNoninative(nounBase, std::get<0>(nounTuple));

				if (NounSpecialPluralFormIsInDictionary(nounNominavitePlural))
				{
					NounRecord nounRecord = GetNounRecordFromDictionary_BySpecialPluralForm(nounNominavitePlural);

					NounStruct ns{ std::get<1>(nounTuple), nounRecord };

					if (!NounScructIsAlreadyInArray(ns, result))
					{
						result.push_back(ns);
					}
				}
			}

		}

		
	
	}


	
	return result;
}

NounDeclencion CalculateNounDeclention(NounRecord nounRecord)
{

	//Xperimental -- need to find if here might be more than 1 declention

	for (auto& i : DeclentionConditionMap)
	{
		if (i.second(nounRecord))
		{
			return i.first;
		}
	}

	std::cout << "Error in CalculateNounDeclention" << std::endl;

	return{};
}


void LoadNounDeclencionCaseTable()
{

	nounDeclencionCaseTable.clear();

#ifdef _WIN32
	std::ifstream f("C:/Workplace/ChineseJournal/rudict/grammar_case.csv");

#else
	std::ifstream f("/home/devuser/workplace/rudict/grammar_case.csv");
#endif

	std::string line;
	std::wstring wline;

	if (f.is_open())
	{

		std::cout << "File found!" << std::endl;

		std::vector<std::pair<NounGrammaticalCase, std::wstring>> currentGrammaticalCaseMap;
		std::wstring currentNounDeclencion;


		while (getline(f, line))
		{
			std::vector<std::string> lineArr;

			boost::split_regex(lineArr, line, boost::regex(";"));

			if (lineArr[0] != "")
			{
				if (currentNounDeclencion == L"")
				{
					currentNounDeclencion = string_to_wstring(lineArr[0]);
				}
				else
				{
					nounDeclencionCaseTable.push_back(NounDeclencionCaseTableRecord{ WStringToNounDeclencion(currentNounDeclencion), currentGrammaticalCaseMap });

					currentNounDeclencion = string_to_wstring(lineArr[0]);

					currentGrammaticalCaseMap.clear();
				}

			}

			currentGrammaticalCaseMap.push_back({ WStringToNounGrammaticalCase(string_to_wstring(lineArr[1])), string_to_wstring(lineArr[2]) });

		}
		//Add last one
		if (currentNounDeclencion != L"")
		{
			nounDeclencionCaseTable.push_back(NounDeclencionCaseTableRecord{ WStringToNounDeclencion(currentNounDeclencion), currentGrammaticalCaseMap });
		}

		f.close();
	}
	else
	{
		std::cout << "file not found!" << std::endl;
	}


}

void LoadFrequentWordSet()
{

	
#ifdef _WIN32
	std::ifstream f("C:/Workplace/ChineseJournal/rudict/frequent_nouns_2000.csv");

#else
	std::ifstream f("/home/devuser/workplace/rudict/frequent_nouns_2000.csv");
#endif

	std::string line;
	std::wstring wline;

	if (f.is_open())
	{
		std::cout << "File found!" << std::endl;
		while (getline(f, line))
		{

			wline = string_to_wstring(line);
			NounRecord nounRecord(wline);

			NounRecordArr.push_back(nounRecord);

		}
		f.close();
	}
	else
	{
		std::cout << "file not found!" << std::endl;
	}


}